//------------------------------------------------------------------------------
// File: AC3Plugin.cpp
//------------------------------------------------------------------------------


#include <assert.h>

#include "source\wav_source.h"
#include "sink\sink_raw.h"
#include "sink\sink_wav.h"
#include "parsers\ac3\ac3_enc.h"
#include "filters\convert.h"
#include "filter_graph.h"
#include "win32\winspk.h"
#include "parsers\spdif\spdif_header.h"
#include "parsers\spdif\spdif_wrapper.h"
#include <mmsystem.h>

#include "AC3Plugin.h"

class PrivateData
{
public:
  AC3Enc      enc;
  FilterChain chain;
  Converter *conv;
  Speakers spk;
  Chunk raw_chunk;
  Chunk ac3_chunk;
  SPDIFWrapper spdifer;

	PrivateData() 
	{
		conv = new Converter( 2048 );
	}
	
	~PrivateData() 
	{
		delete conv;
	}
};


#define GETPRIVATEDATA ((PrivateData*)m_privateData)

AC3Plugin::AC3Plugin()
{ 
	m_privateData = new PrivateData;
}

AC3Plugin::~AC3Plugin()
{
	assert(m_privateData);
	delete m_privateData;
}

// returns 0 if failure, 1 if ok
int AC3Plugin::SetParameters(WAVEFORMATEX *wfx, int bitRate)
{
	if( wfx->nChannels < 1 ||
		wfx->nChannels > 6 ||
		wfx->nSamplesPerSec != 48000 )
		return 0;

	assert(m_privateData);
	delete m_privateData;
	m_privateData = new PrivateData;

	PrivateData *priv = GETPRIVATEDATA;

	priv->chain.add_back(priv->conv, "Converter");
	priv->chain.add_back(&priv->enc,  "Encoder");
	priv->conv->set_format(FORMAT_LINEAR);
	priv->conv->set_order(win_order);
	priv->enc.set_bitrate( bitRate );
	wfx2spk(wfx, priv->spk);

	return 1;
}

// returns 0 on failure, 1 on success

int AC3Plugin::ProcessFlow(unsigned char *inBuffer, int inBytes, unsigned char *outBuffer, int *outBytes, int maxOutBytes)
{
	int total = 0;
	PrivateData *priv = GETPRIVATEDATA;

	*outBytes = total;

	priv->raw_chunk.set_rawdata(priv->spk, inBuffer, inBytes, false, 0, false);
	if (!priv->chain.process(&priv->raw_chunk))
		return 0;

    while (!priv->chain.is_empty())
    {
      if (!priv->chain.get_chunk(&priv->ac3_chunk))
      {
        return 0;
      }
	  if( priv->ac3_chunk.size )
	  {
        if (priv->spdifer.parse_frame(priv->ac3_chunk.rawdata, priv->ac3_chunk.size))
		{
			int size = priv->spdifer.get_rawsize();
			unsigned char *data = priv->spdifer.get_rawdata();

			if( total + size > maxOutBytes )
				return 0;

			memcpy( outBuffer, data, size );
			total += size;
			outBuffer += size;
		}
	  }
	}

	*outBytes = total;

	return 1;
}
