//------------------------------------------------------------------------------
// File: ResamplerPlugin.cpp
//------------------------------------------------------------------------------

/*
 based on Secret Rabbit Code audio resampler v0.1.7
 Copyright (c) Erik de Castro Lopo

 http://www.mega-nerd.com/SRC/index.html
*/

/*
* Modified for ReClock by OGO :)
*
* Update of modified version for ReClock by yesgrey3
* Just changed the previous SCR v0.1.2 files with the new SRC v0.1.7 files
*
*/


#include <assert.h>

#include "ResamplerPlugin.h"
#include	"float_cast.h" // it's needed for working with v0.1.7 (added by yesgrey3)
#include "common.h"
#include "samplerate.h"

class PrivateData
{
public:
	SRC_STATE* m_rs;
	float max_peak; // by yesgrey3
	float peak_limit;

	PrivateData(): m_rs(0), max_peak(1.0f), peak_limit(1.0f) 
	{}
	
	~PrivateData()
	{
		if (m_rs)
			src_delete(m_rs);
	}
};

#define GETPRIVATEDATA ((PrivateData*)m_privateData)

ResamplerPlugin::ResamplerPlugin()
{ 
	m_privateData = 0;

	SetParameters(2, 1);
	SetRate(44100,44100);
}

ResamplerPlugin::~ResamplerPlugin()
{
	if(m_privateData)
	{
		PrivateData* pdata = GETPRIVATEDATA;
		delete pdata;
	}
}


bool ResamplerPlugin::SetParameters(int quality, int channelCount, float peak_limit )
{
	if(m_privateData)
	{
		PrivateData* pdata = GETPRIVATEDATA;
		delete pdata;
	}

	m_privateData = new PrivateData;

	GETPRIVATEDATA->max_peak = peak_limit; 
	GETPRIVATEDATA->peak_limit = peak_limit; 

	m_channelCount = channelCount;
	if (channelCount)
	{
		int error;

		switch(quality)
		{
			case 0:
				GETPRIVATEDATA->m_rs = src_new (SRC_LINEAR, m_channelCount, &error);
				break;

			case 1:
				GETPRIVATEDATA->m_rs = src_new (SRC_ZERO_ORDER_HOLD, m_channelCount, &error);
				break;

			case 2:
				GETPRIVATEDATA->m_rs = src_new (SRC_SINC_FASTEST, m_channelCount, &error);
				break;

			case 3:
				GETPRIVATEDATA->m_rs = src_new (SRC_SINC_MEDIUM_QUALITY, m_channelCount, &error);
				break;

			case 4:
				GETPRIVATEDATA->m_rs = src_new (SRC_SINC_BEST_QUALITY, m_channelCount, &error);
				break;

			default:
				return false;
		}

		return(error == SRC_ERR_NO_ERROR);
	}
	return false;
}

bool ResamplerPlugin::SetRate(int inRate, int outRate)
{
	m_inRate = inRate;
	m_outRate = outRate;

	if (inRate && outRate)
		return (src_is_valid_ratio((double)m_outRate/(double)m_inRate) == SRC_TRUE);
	else
		return false;
}

void ResamplerPlugin::ProcessFlow(float *inBuffer, int inSamples, float *outBuffer, int *outSamples, int maxOutSamples)
{
	*outSamples = 0;

	if (m_channelCount && m_inRate && m_outRate && GETPRIVATEDATA && GETPRIVATEDATA->m_rs)
	{
		SRC_DATA l_data;
		
		l_data.data_in = inBuffer;
		l_data.input_frames = inSamples / m_channelCount;
		l_data.input_frames_used = 0;

		l_data.data_out = outBuffer;
		l_data.output_frames = maxOutSamples / m_channelCount;
		l_data.output_frames_gen = 0;

		l_data.src_ratio = (double)m_outRate/(double)m_inRate;

		l_data.end_of_input = 0;

		int error = src_process (GETPRIVATEDATA->m_rs, &l_data);

		if (error == SRC_ERR_NO_ERROR)
		{
			assert(l_data.input_frames_used == l_data.input_frames);
			*outSamples = l_data.output_frames_gen * m_channelCount;
		}
	}
// by yesgrey3 - start
// if samples > max_peak, normalize using peak_limit/max_peak to avoid clipping
   for(int o = 0; o < *outSamples; ++o)
   {
      if (outBuffer[o] > GETPRIVATEDATA->max_peak)
	  {
		  GETPRIVATEDATA->max_peak = outBuffer[o];
	  }
   }
   if (GETPRIVATEDATA->max_peak > GETPRIVATEDATA->peak_limit)
   {
      float xnorm = GETPRIVATEDATA->peak_limit/GETPRIVATEDATA->max_peak;
      for(int o = 0; o < *outSamples; ++o)
	  {
         outBuffer[o] = outBuffer[o]*xnorm;
	  }
   }
// by yesgrey3 - end
}

void ResamplerPlugin::Flush()
{
	if (GETPRIVATEDATA && GETPRIVATEDATA->m_rs)
		src_reset(GETPRIVATEDATA->m_rs);
}
