//------------------------------------------------------------------------------
// File: TimestretchPlugin.cpp
//------------------------------------------------------------------------------

/*
 based on SoundTouch Audio Library v1.2.1
 Copyright (c) Olli Parviainen 2002-2003

 http://sky.prohosting.com/oparviai/soundtouch/README.html
*/

/*
* Modified for ReClock by OGO :)
*
*/

#include <stdio.h>

#include "TimestretchPlugin.h"
#include "FIFOSampleBuffer.h"

#include "SoundTouch.h"

using namespace soundtouch;

class PrivateData
{
	#define SAMPLE_LEN 0x10000

public:
	SoundTouch* m_st;
	soundtouch::SAMPLETYPE m_temp[SAMPLE_LEN];

	PrivateData(): m_st(0) 
	{}

	~PrivateData() 
	{
		if (m_st)
			delete[] m_st;
	}
};

#define GETPRIVATEDATA ((PrivateData*)m_privateData)

TimestretchPlugin::TimestretchPlugin()
{ 
	m_privateData = new PrivateData;

	SetParameters(1, 44100);
	SetPitch(1.0);
}

TimestretchPlugin::~TimestretchPlugin()
{
	assert(m_privateData);
	delete m_privateData;
}

void TimestretchPlugin::SetParameters(int channelCount, int sampleRate)
{
	assert(m_privateData);
	delete m_privateData;
	m_privateData = new PrivateData;

	m_channelCount = channelCount;
	m_latency = 0;

	if (channelCount)
	{
		if (m_channelCount == 2)
		{
			GETPRIVATEDATA->m_st = new SoundTouch[1];
			GETPRIVATEDATA->m_st[0].setChannels(2);
			GETPRIVATEDATA->m_st[0].setSampleRate(sampleRate);
			GETPRIVATEDATA->m_st[0].setSetting(SETTING_USE_QUICKSEEK, false);
			GETPRIVATEDATA->m_st[0].setSetting(SETTING_USE_AA_FILTER, true);
		}
		else
		{
			GETPRIVATEDATA->m_st = new SoundTouch[m_channelCount];
			for(int i=0;i<m_channelCount;i++)
			{
				GETPRIVATEDATA->m_st[i].setChannels(1);
				GETPRIVATEDATA->m_st[i].setSampleRate(sampleRate);
				GETPRIVATEDATA->m_st[i].setSetting(SETTING_USE_QUICKSEEK, false);
				GETPRIVATEDATA->m_st[i].setSetting(SETTING_USE_AA_FILTER, true);
   			}
		}
	}
}

bool TimestretchPlugin::SetPitch(float pitch)
{
	m_pitch = pitch;

	if (m_channelCount && (pitch > 0.0f))
	{
		if (m_channelCount == 2)
		{
			GETPRIVATEDATA->m_st[0].setTempo(pitch);
		}
		else
		{
			for(int i=0;i<m_channelCount;i++)
				GETPRIVATEDATA->m_st[i].setTempo(pitch);
		}

		Flush();

		return true; // true for success
	}
	else
		return false;
}


void TimestretchPlugin::ProcessFlow(float *inBuffer, int inSamples, float *outBuffer, int *outSamples, int maxOutSamples)
{
	*outSamples = 0;

	if (m_channelCount && (m_pitch > 0.0f) && GETPRIVATEDATA && GETPRIVATEDATA->m_st)
	{
		if (m_channelCount == 2)
		{
			// input
			GETPRIVATEDATA->m_st[0].putSamples(inBuffer, inSamples/2);
			m_latency += ((float)inSamples) / m_pitch;

			// output
			float *osample = outBuffer;
			*outSamples = 0;

			while (maxOutSamples > 0)
			{
				int l_avail = GETPRIVATEDATA->m_st[0].receiveSamples(osample, maxOutSamples/2)*2;
				if (l_avail == 0)
					break;
				m_latency -= l_avail;
				(*outSamples) += l_avail;
				maxOutSamples -= l_avail;
				assert(maxOutSamples >= 0);
			}
		}
		else
		{
			*outSamples = 0;

			for (int c=0;c<m_channelCount;c++)
			{
				// input
				int l_inSamples = inSamples / m_channelCount;
				float* isample = inBuffer+c;

				while (l_inSamples > 0)
				{
					soundtouch::SAMPLETYPE* osample = GETPRIVATEDATA->m_temp;
					
					int i;

					for(i=0; i<SAMPLE_LEN;)
					{
						*osample++ = *isample;
						isample += m_channelCount;
						i++;
						if (--l_inSamples == 0)
							break;
					}

					GETPRIVATEDATA->m_st[c].putSamples(GETPRIVATEDATA->m_temp, i);
					m_latency += ((float)i) / m_pitch;
				}		

				// output
				float *osample = outBuffer+c;

				int l_maxOutSamples = maxOutSamples / m_channelCount;
				while (l_maxOutSamples > 0)
				{
					int l_sampleLen = (SAMPLE_LEN > l_maxOutSamples ? l_maxOutSamples : SAMPLE_LEN);
					int l_avail = GETPRIVATEDATA->m_st[c].receiveSamples(GETPRIVATEDATA->m_temp, l_sampleLen);
					if (l_avail == 0)
						break;
					m_latency -= l_avail;
					(*outSamples) += l_avail;
					l_maxOutSamples -= l_avail;
					assert(l_maxOutSamples >= 0);
					for(int i=0; i<l_avail; i++)
					{
						*osample = GETPRIVATEDATA->m_temp[i];
						osample += m_channelCount;
					}
				}
			}
		}
	}
}

void TimestretchPlugin::Flush()
{
	if (m_channelCount && GETPRIVATEDATA && GETPRIVATEDATA->m_st)
	{
		if (m_channelCount == 2)
		{
			GETPRIVATEDATA->m_st[0].flush();
			while (GETPRIVATEDATA->m_st[0].receiveSamples(SAMPLE_LEN))
				;
		}
		else
		{
			for(int i=0;i<m_channelCount;i++)
			{
				GETPRIVATEDATA->m_st[i].flush();
				while (GETPRIVATEDATA->m_st[i].receiveSamples(SAMPLE_LEN))
					;
			}
		}
	}

	m_latency = 0;
}
